from typing import Union, Optional

def validate_checkdigit(
        number: Union[int, str],
        num_digits: Optional[int]=None,
        weights: Optional[list]=None,
    )-> Optional[Union[int,str]]:
    """ Validate check-digit as mod 11.
    Parameters
    ----------
    number: Union[int, str]
        to validate code
    num_digits: int
        the number of digits or chracters. default is None.
        to use length of number as string.
    weights: Optional[list]
         the weighting factors associated with each digit.
         default is None:
         i.e: to use [6,5,4,3,2] if num_digits is 5
              or length of numbers as string.

    Returns:
        result: return code without checkdigit if code is valid, otherwise None.
    """

    input_as_int = False
    if isinstance(number, int):
        input_as_int = True
        number = str(number)

    for r in ((".", ""),("-","")):
        number = number.replace(*r)

    if num_digits:
        number=number.zfill(num_digits)
    if len(number) != num_digits:
        check_digit = int(number[-1])
        number = number[:-1]
    else:
        return [number, int(number)][input_as_int]

    len_number = len(number)
    if not num_digits:
        num_digits = len_number

    if len_number != num_digits:
        result=None
    else:
        weights = weights or [x for x in range(num_digits + 1, 1, -1)]
        result = sum(w * (int(x)) for w, x in zip(weights, number))
        result = (11 - (result % 11)) == check_digit

    number = [number, int(number)][input_as_int]
    result = number if result else None
    return result

def calc_checkdigit(
        number: Union[int, str],
        num_digits: int=0,
        weights: Optional[list]=None,
        only_checkdigit: bool = False,
    ) -> Optional[Union[int, str]]:
    """ Validate check-digit as mod 11.
    Parameters
    ----------
    number: Union[int, str]
        the number or charaters for generate check digits.
    num_digits: int
        the number of digits or chracters. default is None.
        to use length of number as string.
    weights: Optional[list]
         the weighting factors associated with each digit.
         default is None:
         i.e: to use [6,5,4,3,2] if num_digits is 5
              or length of numbers as string.
    only_checkdigit: bool
         if set ``True``, return only check digit.

    Returns:
        result: Optional[Union[int, str]]
        return code is same as type at input type of `number`.
        input str -> return str, input int -> ireturn int.
    """

    if isinstance(number, int):
        input_as_int = True
        number = str(number)
    elif isinstance(number, str):
        input_as_int = False
    else:
        return None

    for r in ((".", ""),("-","")):
        number = number.replace(*r)

    if number and not num_digits:
        num_digits = len(number)

    if num_digits and len(number) < num_digits:
        number = number.zfill(num_digits)

    weights = weights or [x for x in range(num_digits+1, 1, -1)]
    result = sum(w * (int(x)) for w, x in zip(weights, number))
    checkdigit = str(11 - (result % 11))   # type: ignore
    number = number + checkdigit           # type: ignore
    if input_as_int:
        number = int(number)
        checkdigit = int(checkdigit)       # type: ignore
    if only_checkdigit:
        return checkdigit
    else:
        return number




import numpy as np
import pandas as pd
from typing import List, Optional, Union, Any
from .singledispatchmethod import singledispatchmethod
from .immutable_dict import ImmutableDict

class JpPrefecture(object):
    def __init__(self):
        self.__prefecture_names = {
            "北海道" : 'Hokkaido',
            "青森県" : 'Aomori',
            "岩手県" : 'Iwate',
            "宮城県" : 'Miyagi',
            "秋田県" : 'Akita',
            "山形県" : 'Yamagata',
            "福島県" : 'Fukushima',
            "茨城県" : 'Ibaraki',
            "栃木県" : 'Tochigi',
            "群馬県" : 'Gunma',
            "埼玉県" : 'Saitama',
            "千葉県" : 'Chiba',
            "東京都" : 'Tokyo',
            "神奈川県" : 'Kanagawa',
            "新潟県" : 'Niigata',
            "富山県" : 'Toyama',
            "石川県" : 'Ishikawa',
            "福井県" : 'Fukui',
            "山梨県" : 'Yamanashi',
            "長野県" : 'Nagano',
            "岐阜県" : 'Gifu',
            "静岡県" : 'Shizuoka',
            "愛知県" : 'Aichi',
            "三重県" : 'Mie',
            "滋賀県" : 'Shiga',
            "京都府" : 'Kyoto',
            "大阪府" : 'Osaka',
            "兵庫県" : 'Hyogo',
            "奈良県" : 'Nara',
            "和歌山県" : 'Wakayama',
            "鳥取県" : 'Tottori',
            "島根県" : 'Shimane',
            "岡山県" : 'Okayama',
            "広島県" : 'Hiroshima',
            "山口県" : 'Yamaguchi',
            "徳島県" : 'Tokushima',
            "香川県" : 'Kagawa',
            "愛媛県" : 'Ehime',
            "高知県" : 'Kochi',
            "福岡県" : 'Fukuoka',
            "佐賀県" : 'Saga',
            "長崎県" : 'Nagasaki',
            "熊本県" : 'Kumamoto',
            "大分県" : 'Oita',
            "宮崎県" : 'Miyazaki',
            "鹿児島県" : 'Kagoshima',
            "沖縄県" : 'Okinawa',
        }

        # Index is code (JIS X 0401-1973)
        self.prefectures = pd.DataFrame(
            dict(
                name = [p for p in self.__prefecture_names.keys()],
                short_name = [p[:-1] for p in self.__prefecture_names.keys()],
                alphabet_name = [p for p in self.__prefecture_names.values()],
            ),
            index = pd.Index(range(1, 1 + len(self.__prefecture_names)),
                             name="code"),
        )
        del self.__prefecture_names

        self.__code2name = ImmutableDict({
            code: name
            for name, code in zip(self.prefectures.name,
                                  self.prefectures.index)
        })

        self.__code2alphabet = ImmutableDict({
            code: alphabet
            for alphabet, code in zip(self.prefectures.alphabet_name,
                                  self.prefectures.index)
        })

        self.__name2code = ImmutableDict({
            **{name: code
               for name, code in zip(self.prefectures.name,
                                     self.prefectures.index)},
            **{name: code
               for name, code in zip(self.prefectures.short_name,
                                     self.prefectures.index)},
            **{name: code
               for name, code in zip(self.prefectures.alphabet_name,
                                     self.prefectures.index)},
        })

        self.__alphabet2name = ImmutableDict({
            **{alphabet: name
            for alphabet, name in zip(self.prefectures.alphabet_name,
                                      self.prefectures.name)},
            **{name: code
               for name, code in zip(self.prefectures.short_name,
                                     self.prefectures.name)},
            **{alphabet: name
            for alphabet, name in zip(self.prefectures.name,
                                      self.prefectures.name)},
        })

        self.__name2alphabet = ImmutableDict({
            **{name: alphabet
              for name, alphabet in zip(self.prefectures.name,
                                        self.prefectures.alphabet_name)},
            **{name: alphabet
              for name, alphabet in zip(self.prefectures.short_name,
                                        self.prefectures.alphabet_name)},
            **{name: alphabet
              for name, alphabet in zip(self.prefectures.alphabet_name,
                                        self.prefectures.alphabet_name)},
    })

    @singledispatchmethod
    def name2code(self, arg: Any) -> Union[list,Optional[int],pd.Series]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @name2code.register(type(None))
    def _name2code_none(self,
            name: None,
            ignore_case: bool=False,
        ):
        """ Catch None and return None """
        return None

    @name2code.register(str)
    def _name2code_str(self,
            name: str,
            ignore_case: bool=False,
        ) -> Optional[int]:
        """ Convert prefecture name to code """
        try:
            name = [name, name.capitalize()][ignore_case]
            code = self.__name2code[name]
        except KeyError:
            code = None
        return code

    @name2code.register(list)
    def _name2code_list(self,
            name_list: List[str],
            ignore_case: bool=False,
        ) -> list:
        """ Convert list of prefecture name to code """
        code = [self.name2code(x, ignore_case) for x in name_list]
        return code

    @name2code.register(pd.Series)
    def _name2code_series(self,
            name_series: pd.Series,
            ignore_case: bool=False,
        ) -> pd.Series:
        """ Convert pandas series of prefecture name to code """
        try:
            name_series = [name_series,
                           name_series.str.capitalize()][ignore_case]
            code = name_series.map(self.__name2code)
        except KeyError:
            code = pd.Series([])
        return code

    @singledispatchmethod
    def code2name(self, arg: Any) -> Union[Optional[str],list,pd.Series]:
        """ Convert prefecture code to name """
        raise TypeError('Unsupport Type')

    @code2name.register(type(None))
    def _code2name_none(self,
            codeL: None,
            ascii: bool=False,
        ):
        """ Catch None and return None """
        return None

    @code2name.register(int)
    def _code2name_int(self,
            code: int,
            ascii: bool=False,
        ) -> Optional[str]:
        """ Convert prefecture code to name """
        try:
            name = [self.__code2name[code],
                    self.__code2alphabet[code]][ascii]
        except KeyError:
            name = None
        return name

    @code2name.register(str)
    def _code2name_str(self,
            code: str,
            ascii: bool=False,
        ) -> Optional[str]:
        """ Convert prefecture code to name """
        try:
            name = [self.__code2name[int(code)],
                    self.__code2alphabet[int(code)]][ascii]
        except KeyError:
            name = None
        return name

    @code2name.register(list)
    def _code2name_list(self,
            code_list: list,
            ascii: bool=False,
        ) -> list:
        """ Convert list of prefecture code to name """
        name = [self.code2name(x, ascii) for x in code_list]
        return name

    @code2name.register(pd.Series)
    def _code2name_series(self,
            code_series: pd.Series,
            ascii: bool=False,
        ) -> pd.Series:
        """ Convert pandas series of prefecture code to alphabet_name """
        try:
            code_series = code_series.astype(int)
            name = [ code_series.map(self.__code2name),
                     code_series.map(self.__code2alphabet) ][ascii]
        except KeyError:
            name = pd.Series([])
        return name


    @singledispatchmethod
    def name2normalize(self, arg: Any) -> Union[Optional[str],list,pd.Series]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @name2normalize.register(type(None))
    def _name2normalize_none(self,
            name: None,
            ascii: bool=False,
            ignore_case: bool=False
        ):
        """ Catch None and return None """
        return None

    @name2normalize.register(str)
    def _name2normalize_str(self,
            name: str,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> Optional[str]:
        """ Convert prefecture name to name or alphabet_name """
        try:
            name = [name, name.capitalize()][ignore_case]
            normalize = [ self.__alphabet2name[name],
                     self.__name2alphabet[name] ][ascii]
        except KeyError:
            normalize = None
        return normalize

    @name2normalize.register(list)
    def _name2normalize_list(self,
            name_list: list,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> list:
        """ Convert list of prefecture name to alphabet_name """
        code = [self.name2normalize(x, ascii, ignore_case) for x in name_list]
        return code

    @name2normalize.register(pd.Series)
    def _name2normalize_series(self,
            name_series: pd.Series,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> pd.Series:
        """ Convert pandas series of prefecture name to alphabet_name """
        try:
            name_series = [name_series,
                           name_series.str.capitalize()][ignore_case]
            code = [ name_series.map(self.__alphabet2name),
                     name_series.map(self.__name2alphabet) ][ascii]
        except KeyError:
            code = pd.Series([])
        return code


    @singledispatchmethod
    def validate(self, arg: Any) -> Optional[Union[bool,List[bool],pd.Series]]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @validate.register(type(None))
    def _validate_none(self,
            name: None,
            ignore_case: bool=False
        ) -> Union[bool,List[bool],pd.Series]:
        """ Catch None and return None """
        return None

    @validate.register(str)
    def _validate_str(self,
            name: str,
            ignore_case: bool=False
        ) -> bool:
        """ validate a prefecture name """
        try:
            name = [name, name.capitalize()][ignore_case]
            v = name in self.__name2code.keys()
        except:
            v = False
        return v

    @validate.register(list)
    def _validate_list(self,
            name_list: List[str],
            ignore_case: bool=False
        ) -> list:
        """ validate list of prefecture name """
        try:
            v = [ self.validate(x, ignore_case) for x in name_list]
        except:
            v = [False]
        return v

    @validate.register(pd.Series)
    def _validate_series(self,
            name_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.Series:
        """ validate pandas series of prefecture name """
        try:
            v = [ self.validate(x, ignore_case) for x in np.asarray(name_series)]
        except KeyError:
            v = [False]
        return pd.Series(v)

jp_prefectures = JpPrefecture()



try:
    # python 3.8 or later
    from functools import singledispatchmethod
except ImportError:
    from functools import singledispatch, update_wrapper

    def singledispatchmethod(func):  # type: ignore
        dispatcher = singledispatch(func)
        def wrapper(*args, **kw):
            return dispatcher.dispatch(args[1].__class__)(*args, **kw)
        wrapper.register = dispatcher.register
        update_wrapper(wrapper, func)
        return wrapper



import numpy as np
import pandas as pd
import re
from typing import Optional, Union, Any
from pathlib import Path
from .singledispatchmethod import singledispatchmethod
from .immutable_dict import ImmutableDict
from .checkdigit import calc_checkdigit, validate_checkdigit
from .jp_prefecture import JpPrefecture
from .utils import is_alpha

class JpCity(JpPrefecture):
    def __init__(self):
        # City Code: JIS X 0402

        super().__init__()
        self.cities = pd.read_csv(Path(__file__).parent / 'data/cities.csv',
                                  index_col = 0)

        self.cities['prefCode'] = pd.to_numeric(self.cities['prefCode'],
                                                downcast='integer')
        self.cities['cityCode'] = pd.to_numeric(self.cities['cityCode'],
                                                downcast='integer')
        self.cities['latitude'].astype(float)
        self.cities['longitude'].astype(float)
        self.cities['bigCityFlag'] = pd.to_numeric(self.cities['bigCityFlag'],
                                                downcast='integer')
        self.__citycode2name = ImmutableDict({
            code: name
            for code, name in zip(self.cities.cityCode,
                                  self.cities.cityName)
        })

        self.__citycode2alphabet = ImmutableDict({
            code: alphabet
            for code, alphabet in zip(self.cities.cityCode,
                                  self.cities.cityAlphabet)
        })

        self.__cityname2code = ImmutableDict({
            **{name: code
               for name, code in zip(self.cities.cityName,
                                     self.cities.cityCode)},
            **{name: code
               for name, code in zip(self.cities.cityAlphabet,
                                     self.cities.cityCode)},
            **{name.title(): code
               for name, code in zip(self.cities.cityAlphabet,
                                     self.cities.cityCode)},
        })

        self.__cityname2alphabet = ImmutableDict({
            **{name: alphabet
               for name, alphabet in zip(self.cities.cityName,
                                     self.cities.cityAlphabet)},
            **{name: alphabet
               for name, alphabet in zip(self.cities.cityAlphabet,
                                     self.cities.cityAlphabet)},
            **{name.title(): alphabet
               for name, alphabet in zip(self.cities.cityAlphabet,
                                     self.cities.cityAlphabet)},
        })

        self.__cityalphabet2name = ImmutableDict({
            **{alphabet: name
               for alphabet, name in zip(self.cities.cityName,
                                     self.cities.cityName)},
            **{alphabet: name
               for alphabet, name in zip(self.cities.cityAlphabet,
                                     self.cities.cityName)},
            **{alphabet.title(): name
               for alphabet, name in zip(self.cities.cityAlphabet,
                                     self.cities.cityName)},
        })

    @singledispatchmethod
    def cityname2code(self,
            arg: Any
        ) -> Union[Optional[Union[int,str]],list,pd.Series]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @cityname2code.register(type(None))
    def _cityname2code_none(self,
            name: str,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> Optional[Union[int,str]]:
        """ Catch None and return None """
        return None

    @cityname2code.register(str)
    def _cityname2code_str(self,
            name: str,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> Optional[str]:
        """ Convert cityName to cityCode """
        try:
            name = [name, name.title()][ignore_case]
            code = self.__cityname2code[name]
            if code:
                code = [code, str(code).zfill(5)][as_str]
                code = [code, calc_checkdigit(code)][checkdigit]
        except KeyError:
            code = None
        return code

    @cityname2code.register(re.Pattern)
    def _cityname2code_str(self,
            name: re.Pattern,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> list:
        """ Convert cityName to cityCode """
        cities = [ re.search(name, x) for x in self.__cityname2code.keys() ]
        codes = list()
        for x in cities:
            if x:
                code = self.__cityname2code[x.group(0)]
                if code not in codes:
                    codes.append(code)
        return codes

    @cityname2code.register(list)
    def _cityname2code_list(self,
            name_list: list,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> list:
        """ Convert list of cityName to cityCode """
        code = [self.cityname2code(x, ignore_case, checkdigit)
                                   for x in name_list]
        return code

    @cityname2code.register(pd.Series)
    def _cityname2code_series(self,
            name_series: pd.Series,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> pd.Series:
        """ Convert pandas series of cityName to cityCode """
        try:
            name_series = [name_series,
                           name_series.str.title()][ignore_case]
            code = name_series.map(self.__cityname2code)
            code = [code,
                    code.map(calc_checkdigit)][checkdigit]
        except KeyError:
            code = pd.Series([])
        return code


    @singledispatchmethod
    def citycode2name(self, arg: Any) -> Union[Optional[str],list,pd.Series]:
        """ Convert cityCode to cityName """
        raise TypeError('Unsupport Type')

    @citycode2name.register(type(None))
    def _citycode2name_none(self,
            code: None,
            ascii: bool=False,
        ):
        return None

    @citycode2name.register(int)
    def _citycode2name_int(self,
            code: int,
            ascii: bool=False,
        ) -> Optional[str]:
        """ Convert cityCode to cityName
            if set ``True`` to ascii, return cityname as alphabet_name
        """
        if len(str(code)) == 6:
            code = validate_checkdigit(code)  # type: ignore
        try:
            name = [ self.__citycode2name[code],
                     self.__citycode2alphabet[code]][ascii]
        except KeyError:
            name = None
        return name

    @citycode2name.register(str)
    def _citycode2name_str(self,
            code: str,
            ascii: bool=False,
        ) -> Optional[str]:
        """ Convert cityCode to cityName
            if set ``True`` to ascii, return cityname as alphabet_name
        """
        if len(code) == 6:
            code = validate_checkdigit(int(code), 5)  # type: ignore
        else:
            code = int(code)   # type: ignore
        try:
            name = [ self.__citycode2name[code],
                     self.__citycode2alphabet[code]][ascii]
        except:
            name = None
        return name

    @citycode2name.register(list)
    def _citycode2name_list(self,
            code_list: list,
            ascii: bool=False,
        ) -> list:
        """ Convert list of cityCode to CityName
            if set ``True`` to ascii, return cityname as alphabet_name
        """
        name = [self.citycode2name(x, ascii) for x in code_list]
        return name

    @citycode2name.register(pd.Series)
    def _citycode2name_series(self,
            code_series: pd.Series,
            ascii: bool=False,
        ) -> pd.Series:
        """ Convert pandas series of cityCode to cityName
            if set ``True`` to ascii, return cityname as alphabet_name
        """
        try:
            code_series.astype(int)
            name = [ code_series.map(self.__citycode2name),
                     code_series.map(self.__citycode2alphabet)][ascii]
        except:
            name = pd.Series([])
        return name

    @singledispatchmethod
    def cityname2normalize(self, arg: Any
        ) -> Union[Optional[int],list,pd.Series]:
        """ Normalize cityName
            if set ``True`` to ascii, return cityName as alphabet_name
        """
        raise TypeError('Unsupport Type')

    @cityname2normalize.register(type(None))
    def _cityname2normalize_none(self,
            name: None,
            ascii: bool=False,
            ignore_case: bool=False
        ):
        return None

    @cityname2normalize.register(str)
    def _cityname2normalize_str(self,
            name: str,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> Optional[str]:
        """ Normalize cityName
            if set ``True`` to ascii, return cityName as alphabet_name
        """
        try:
            name = [name, name.title()][ignore_case]
            name = [ self.__cityalphabet2name[name],
                     self.__cityname2alphabet[name] ][ascii]
        except KeyError:
            name = None    # type: ignore
        return name

    @cityname2normalize.register(list)
    def _cityname2normalize_list(self,
            name_list: list,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> list:
        """ Convert list of cityName to cityCode
            if set ``True`` to ascii, return cityName as alphabet_name
        """
        name = [self.cityname2normalize(x,ascii,ignore_case)
                                        for x in name_list]
        return name

    @cityname2normalize.register(pd.Series)
    def _cityname2normalize_series(self,
            name_series: pd.Series,
            checkdigit=False,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> pd.Series:
        """ Convert pandas series of cityName to cityCode
            if set ``True`` to ascii, return cityName as alphabet_name
        """
        try:
            name_series = [name_series,
                           name_series.str.title()][ignore_case]
            name = [ name_series.map(self.__cityalphabet2name),
                     name_series.map(self.__cityname2alphabet) ][ascii]
        except KeyError:
            name = pd.Series([])
        return name

    def get_prefcode(self, citycode: Union[int, str])-> Optional[int]:
        if isinstance(citycode, int):
            citycode = str(citycode)
        try:
            return int(citycode[:2])
        except:
            return None

    def citycode2normalize(self,
            citycode: Union[int, str],
            as_str: bool=False,
        )-> Optional[int]:
        if isinstance(citycode, int):
            citycode = str(citycode).zfill(5)

        try:
            if len(citycode) == 6:
                citycode = validate_checkdigit(citycode)  # type: ignore
            citycode = [int(citycode), citycode][as_str]
        except:
            citycode = None

        return citycode

    def findcity(self,
            name: re.Pattern,
            ignore_case: bool=False,
            ascii: bool=False,
        ) -> list:
        flag = [0, re.IGNORECASE][ignore_case]
        result = [ re.search(name, x, flag)
                   for x in self.__cityname2code.keys() ]
        cities = list()
        for x in result:
            if x:
                city = x.group(0).title()
                if city in self.__cityname2code.keys():
                    name = [ city, self.__cityname2alphabet[city] ][ascii]
                    if name not in cities:
                        cities.append(name)

        return cities


    @singledispatchmethod
    def cityname2prefcode(self, arg: Any,
        ) -> Union[Optional[int],list,pd.Series]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @cityname2prefcode.register(type(None))
    def _cityname2prefcode_none(self,
            name: None,
            ignore_case: bool=False
        ):
        """ Catch None and return None """
        return None

    @cityname2prefcode.register(str)
    def _cityname2prefcode_str(self,
            name: str,
            ignore_case: bool=False
        ) -> Optional[int]:
        """ Convert CityName to Prefecture Code"""
        name = [name, name.title()][ignore_case]
        code = self.__cityname2code[name]
        prefcode = self.get_prefcode(code)
        return prefcode

    @cityname2prefcode.register(list)
    def _cityname2prefcode_list(self,
            name_list: list,
            ignore_case: bool=False
        ) -> list:
        """ Convert list of cityName to Prefecture Code """
        code = [self.cityname2prefcode(x, ignore_case) for x in name_list]
        return code

    @cityname2prefcode.register(pd.Series)
    def _cityname2prefcode_series(self,
            name_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.Series:
        """ Convert pandas series of cityName to Prefecture Code """
        try:
            name_series = [name_series,
                           name_series.str.title()][ignore_case]
            code = name_series.map(self.__cityname2code)
            code = code.map(self.get_prefcode)
        except KeyError:
            code = pd.Series([None])
        return code

    @singledispatchmethod
    def cityname2prefecture(self, arg: Any) -> Union[Optional[str],list,pd.Series]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @cityname2prefecture.register(type(None))
    def _cityname2prefecture_none(self,
            name: None,
            ascii: bool=False,
            ignore_case: bool=False
        ):
        """ Catch None and return None """
        return None

    @cityname2prefecture.register(str)
    def _cityname2prefecture_str(self,
            name: str,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> Optional[str]:
        """ Convert CityName to Prefecture Name"""
        name = [name, name.title()][ignore_case]
        code = self.cityname2prefcode(name)
        name = code and self.code2name(code, ascii)  # type: ignore
        return name

    @cityname2prefecture.register(list)
    def _cityname2prefecture_list(self,
            name_list: list,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> list:
        """ Convert list of CityName to Prefecture Name"""
        name = [self.cityname2prefecture(x, ascii,ignore_case)
                                         for x in name_list]
        return name

    @cityname2prefecture.register(pd.Series)
    def _cityname2prefecture_series(self,
            name_series: pd.Series,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> pd.Series:
        """ Convert pandas series of cityName to Prefecture Name """
        try:
            name_series = [name_series,
                           name_series.str.title()][ignore_case]
            code = name_series.map(self.cityname2prefcode)
            name = code.apply(self.code2name, ascii=ascii)
        except KeyError:
            # v = [None]
            name = pd.Series([None])
        # return pd.Series(v)
        return name


    @singledispatchmethod
    def cityname2geodetic(self,
            arg: Any,
        ) -> Union[Optional[tuple],list,pd.DataFrame]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @cityname2geodetic.register(str)
    def _cityname2geodetic_str(self,
            name: str,
            ignore_case: bool=False
        ) -> Optional[tuple]:
        """ Return Latitude and Longitude of CityName """
        code = self.cityname2code(name, ignore_case)
        if code:
            city = self.cities.loc[self.cities['cityCode'] == code]
            geodetic = ( city['latitude'].values[0],
                         city['longitude'].values[0] )
        else:
            geodetic = None    # type: ignore
        return geodetic

    @cityname2geodetic.register(list)
    def _cityname2geodetic_list(self,
            name_list: list,
            ignore_case: bool=False
        ) -> list:
        """ Return Latitude and Longitude of CityName """
        code = self.cityname2code(name_list, ignore_case)
        geodetic = list()
        for c in code:
            city = self.cities.loc[self.cities['cityCode'] == c]
            pos = ( city['latitude'].values[0],
                    city['longitude'].values[0] )
            geodetic.append(pos)
        if len(geodetic) == 0:
            geodetic = None    # type: ignore
        return geodetic

    @cityname2geodetic.register(pd.Series)
    def _cityname2geodetic_series(self,
            name_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.DataFrame:
        """ Return Latitude and Longitude of CityName """
        code = name_series.apply(self.cityname2code, ignore_case)
        geodetic = self.cities.loc[self.cities['cityCode'].isin(code.values),
                                   ['latitude','longitude']]
        geodetic.insert(0, 'cityName', name_series.values )
        return geodetic

    @singledispatchmethod
    def citycode2geodetic(self,
            arg: Any,
        ) -> Union[Optional[tuple],list,pd.DataFrame]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @citycode2geodetic.register(str)
    @citycode2geodetic.register(int)
    def _citycode2geodetic(self,
            code: Union[int,str],
            ignore_case: bool=False
        ) -> Optional[tuple]:
        """ Return Latitude and Longitude of CityName """

        citycode = self.citycode2normalize(code)
        if citycode:
            city = self.cities.loc[self.cities['cityCode'] == citycode]
            geodetic = ( city['latitude'].values[0],
                         city['longitude'].values[0] )
        else:
            geodetic = None    # type: ignore
        return geodetic

    @citycode2geodetic.register(list)
    def _citycode2geodetic_list(self,
            code_list: list,
            ignore_case: bool=False
        ) -> list:
        """ Return Latitude and Longitude of CityName """
        city_codes = [ self.citycode2normalize(x) for x in code_list ]
        geodetic = list()
        try:
            for c in city_codes:
                city = self.cities.loc[self.cities['cityCode'] == c]
                pos = ( city['latitude'].values[0],
                        city['longitude'].values[0] )
                geodetic.append(pos)
        except:
            geodetic = list()

        if len(geodetic) == 0:
            geodetic = None    # type: ignore
        return geodetic

    @citycode2geodetic.register(pd.Series)
    def _citycodegeodetic_series(self,
            code_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.DataFrame:
        """ Return Latitude and Longitude of CityName """
        city_codes =  code_series.map(self.citycode2normalize)
        geodetic = self.cities.loc[( self.cities['cityCode']
                                         .isin(city_codes.values)),
                                     ['cityCode','latitude','longitude']]
        return geodetic

    @singledispatchmethod
    def validate_city(self, arg: Any) -> Optional[Union[bool,list,pd.Series]]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @validate_city.register(type(None))
    def _validate_city_none(self,
            name: None,
            ignore_case: bool=False
        ) -> bool:
        """ Catch None and return None """
        return False

    @validate_city.register(str)
    def _validate_city_str(self,
            name: str,
            ignore_case: bool=False
        ) -> bool:
        """ validate_city a cityName """
        try:
            name = [name, name.title()][ignore_case]
            v = name in self.__cityname2code.keys()
        except:
            v = False
        return v

    @validate_city.register(list)
    def _validate_city_list(self,
            name_list: list,
            ignore_case: bool=False
        ) -> list:
        """ validate_city list of cityName """
        try:
            v = [ self.validate_city(x,ignore_case) for x in name_list]
        except:
            v = [False]
        return v

    @validate_city.register(pd.Series)
    def _validate_city_series(self,
            name_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.Series:
        """ validate_city pandas series of cityName """
        try:
            v = [ self.validate_city(x,ignore_case)
                                     for x in np.asarray(name_series)]
        except KeyError:
            v = [False]
        return pd.Series(v)

jp_cities = JpCity()



class ImmutableDict(dict):
    def __missing__(self, key):
        return None

    def __setitem__(self, key, value):
        raise TypeError("%r object does not support item assignment" % type(self).__name__)

    def __delitem__(self, key):
        raise TypeError("%r object does not support item deletion" % type(self).__name__)

    def __getattribute__(self, attribute):
        if attribute in ('clear', 'update', 'pop', 'popitem', 'setdefault'):
            raise AttributeError("%r object has no attribute %r" % (type(self).__name__, attribute))
        return dict.__getattribute__(self, attribute)

    def __hash__(self):
        return hash(tuple(sorted(self.iteritems())))

    def fromkeys(self, S, v):
        return type(self)(dict(self).fromkeys(S, v))



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

def get_version(rel_path):
    for line in (this_directory / rel_path).read_text().splitlines():
        if line.startswith('__VERSION__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

setup(
    name="jp_prefecture",
    version=get_version('jp_prefecture/versions.py'),
    license="MIT",
    install_requirements=["pandas"],
    author="iisaka51",
    author_email="iisaka51@gmail.com",
    url="https://github.com/iisaka51/jp_prefecture",
    description="Simple utilitiy convert name of japanese prefectures",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    package_data={'': [ 'data/cities.csv' ]},
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
    ],
)



import pandas as pd
from typing import Union

def is_alpha(word: str)-> bool:
    """ Check word is alphabet.
    Parameters
    ----------
    word: str
        any string

    Returns
    -------
    validate result: bool
        if all characters of word, return ``True`` otherwise return ``False``
    """
    try:
        return word.encode('ascii').isalpha()
    except:
        return False

def is_alnum(word: str)-> bool:
    """ Check word is alphabet and digits.
    Parameters
    ----------
    word: str
        any string

    Returns
    -------
    validate result: bool
        if all characters of word, return ``True`` otherwise return ``False``
    """
    try:
        return word.encode('ascii').isalnum()
    except:
        return False



def df_compare(
        df1: pd.DataFrame,
        df2: pd.DataFrame,
        diff_count: bool=False
    ) -> Union[int,bool]:
    """ Compare DataFrame
    Parameters
    ----------
    df1: pd.DataFrame, df2: pd.DataFrame
        any DataFrame to compare
    diff_count: bool
        if set to ``True`, Return counts of diffs.

    Returns
    -------
    validate result: Union[bool,int]
    """

    diff_df = pd.concat([df1,df2]).drop_duplicates(keep=False)
    diffs = len(diff_df)
    result = [diffs == 0, diffs][diff_count]
    return result
